__all__ = [
    "SpdxLicense",
    "spdx_licenses",
    "detect_license",
]

from yalm._resources import ResourceLoader as _ResourceLoader, \
    KeyedDocument as _KeyedDocument
from yalm._worker import SyncWorker as _SyncWorker, AsyncWorker as _AsyncWorker

from yalm.template.template import Node as _Node
import regex as _re
from typing import Union as _Union
from yalm.template import normalizer as _normalizer, template as _template
from yalm.words import split_and_normalize as _split_and_normalize, \
    is_subset_of as _is_subset_of
from collections import OrderedDict as _OrderedDict

class SpdxLicense:
  def __init__(self, identifier: str, name: str, priority: float, resources: _ResourceLoader) -> None:
    self.id = identifier
    self.name = name
    self.priority = priority
    self._resources = resources

  @property
  def template(self) -> _Node:
    return self._resources.get_template(self.id)

  @property
  def regex(self) -> _re.Pattern:
    return self._resources.get_regex(self.id)

  @property
  def words(self) -> list[str]:
    return self._resources.get_words(self.id)

  @property
  def positive_samples(self) -> list[_KeyedDocument]:
    return self._resources.get_positive_samples(self.id)

  def positive_sample(self, key) -> _KeyedDocument:
    return self._resources.get_positive_sample(self.id, key)

  def _test_words(self, words: list[str], words_sorted: bool=False) -> _Union[float, bool]:
    return _is_subset_of(self.words, words, x_sorted=True, y_sorted=words_sorted)

  def _test_regex(self, text: str, timeout: float = None) -> bool:
    return self.regex.fullmatch(text, timeout=timeout) is not None

class LicenseMatch:
  def __init__(self, text, template):
    self.text = text
    self.template = template

_resources = _ResourceLoader()
_normalizer = _normalizer.TemplateNormalizer(_resources.equivalent_words)

# resource loading
spdx_licenses: _OrderedDict[str, SpdxLicense] = {
  item['id']: SpdxLicense(item['id'], item['name'], item['priority'], _resources) for item in _resources.index
}


def _normalize_text(text: str) -> str:
  return _normalizer(_template.TextNode(text)).text

def _do_regex_test(text: str, normalized_text: str, license_id: str, timeout: float):
  license = spdx_licenses[license_id]
  try:
    if license._test_regex(normalized_text, timeout=timeout):
      return LicenseMatch(text, license)
  except TimeoutError:
    return None

def detect_license(text: str, timeout: float = 2, num_workers: int = 1, best_guess=True) -> LicenseMatch:
  words = _split_and_normalize(text, _resources.equivalent_words, sort=True)
  normalized_text = _normalize_text(text)

  if num_workers == 1:
    worker_raw = _SyncWorker(early_return=best_guess)
  else:
    worker_raw = _AsyncWorker(early_return=best_guess, num_workers=num_workers)

  with worker_raw as worker:
    for license in spdx_licenses.values():
      if license._test_words(words, words_sorted=True):
        # request the worker pool to process regex matching
        worker.request_process(_do_regex_test, text, normalized_text, license.id, timeout)
    return worker.get_results()



__all__ = [
    'Worker',
    'SyncWorker',
    'AsyncWorker',
]

import time
from multiprocessing import Pool, TimeoutError
from abc import ABC, abstractmethod


class Worker:
  @abstractmethod
  def request_process(self, f, *args, **kwargs):
    pass

  @abstractmethod
  def get_results(self):
    pass

  def close(self):
    pass

  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    self.close()


class SyncWorker(Worker):
  def __init__(self, early_return: bool = True):
    self.early_return = early_return
    if self.early_return:
      self.result = None
    else:
      self.results = []

  def request_process(self, f, *args, **kwargs):
    if self.early_return and self.result is not None:
      return
    result = f(*args, **kwargs)
    if result is not None:
      if self.early_return:
        self.result = result
      else:
        self.results.append(result)

  def get_results(self):
    if self.early_return:
      return self.result
    else:
      return self.results


class AsyncWorker(Worker):
  def __init__(self, num_workers=None, timeout=None, early_return: bool = True):
    self.pool = Pool(num_workers)
    self.timeout = timeout
    self.results = []
    self.early_return = early_return

  def request_process(self, f, *args, **kwargs):
    self.results.append(self.pool.apply_async(f, args, kwargs))

  def get_results(self):
    begin_time = time.time()
    unpacked_results = []
    for result in self.results:
      try:
        if self.timeout is None:
          unpacked_result = result.get()
        else:
          rem = max(0, self.timeout - (time.time() - begin_time))
          unpacked_result = result.get(rem)
        if unpacked_result is not None:
          if self.early_return:
            return unpacked_result
          unpacked_results.append(unpacked_result)
      except TimeoutError:
        pass
    return None if self.early_return else unpacked_result

  def close(self):
    self.pool.terminate()
    self.pool.join()



from yalm import licenses

mit, apache, gpl2, gpl3, *_ = list(licenses.spdx_licenses.values())

template = gpl3
key = '9409774ee89c484fe0f4b65d12bda1d2923d4687686858679172d473cd83038c.txt'
key = 'c65379247414a47f9353192b1e85bbcccb2b703ee3f0df95e0d7eb682035f09d.txt'

raw_doc = template.positive_sample(key).document
doc = licenses._normalize_text(raw_doc)
pattern = apache.regex.pattern
